/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File legasm.c */
/* This file contains the legasm interpreter. */

#include <stdio.h>
#include <stdlib.h>

#include "lgd0.h"
#include "config.h"
#include "lgdlib.h"
#include "legasm.h"

#define LVM_RAM_BASE 0x80000000 /* Start of RAM. */
#define PC_NUM 0xff /* Which register is the program counter. */
#define SP_NUM 0xfe /* Which register is the stack pointer. */
#define FL_NUM 0xfd /* Which register is the flags register. */

/* The LVM registers. */
static lvmw_t regs[256];
/* The LVM RAM. */
static lvmw_t *mem_ram;
/* The size of the LVM RAM. */
static size_t mem_ram_size; /* XXX - What type should we use? */
/* The LVM stack area, in reverse order. */
static lvmw_t *mem_stack;
/* The size of the LVM stack area. */
static size_t mem_stack_size;
/* The size effectively malloc()ated for the LVM stack area. */
static size_t mem_stack_all_size;
/* The size of the program. */
static size_t mem_prg_size;
/* The program cache. */
static lvmw_t *mem_cache;
/* The size of the program cache. */
static size_t mem_cache_size;
/* The address of beginning the program cache. */
static lvmw_t mem_cache_addr;

/* The program file being executed. */
static lgd_filedesc_t prgfile;

/* The program is running. */
static char running;

void
legasm_init(char *name)
     /* Initialize the legasm interpreter: this function must be
      * called before any other legasm-related call is performed. The
      * parameter name is the name of the compiled program file to
      * read. */
{
  int i;
  lgd_open(1,&prgfile,name,OPEN_READONLY);
  mem_ram_size=4096;
  mem_ram=lgd_malloc(1,mem_ram_size*sizeof(lvmw_t));
  mem_stack_size=0;
  mem_stack_all_size=0;
  mem_stack=lgd_malloc(1,mem_stack_all_size*sizeof(lvmw_t));
  mem_prg_size=lgd_size(1,prgfile)/sizeof(lvmw_t);
  mem_cache_size=0;
  mem_cache=lgd_malloc(1,mem_cache_size*sizeof(lvmw_t));
  for (i=0;i<0x100;i++) regs[i]=0;
  regs[0x00]=0;
  regs[0x01]=1;
  regs[0x02]=0x80000000;
  regs[0x03]=(lvmw_t)-1; /* Or is ~0 better? */
}

static lvmw_t
readmem(lvmw_t addr)
     /* Read the word at the indicated address in the LVM address
      * space. */
{
  if ((addr>=mem_cache_addr)&&(addr<mem_cache_addr+mem_cache_size))
    return mem_cache[addr-mem_cache_addr];
  if (addr<mem_prg_size) {
#if 0 /* Set to 1 to disable the cache */
    lvmw_t w;
    lgd_read_littleendian_int32(1,prgfile,&w,addr,1);
    return w;
#else
    mem_cache_addr=addr&2047;
    mem_cache_size=2048;
    if (mem_cache_size>mem_prg_size-mem_cache_addr)
      mem_cache_size=mem_prg_size-mem_cache_addr;
    lgd_free(mem_cache);
    mem_cache=lgd_malloc(1,mem_cache_size*sizeof(lvmw_t));
    lgd_read_littleendian_int32(1,prgfile,mem_cache,mem_cache_addr,
				mem_cache_size);
    return mem_cache[addr-mem_cache_addr];
#endif
  }
  if ((addr>=LVM_RAM_BASE)&&(addr<LVM_RAM_BASE+mem_ram_size))
    return mem_ram[addr-LVM_RAM_BASE];
  if (~addr<mem_stack_size)
    return mem_stack[~addr];
  return 0;
}

static void
writemem(lvmw_t addr,lvmw_t data)
     /* Write a word at the indicated address in the LVM address
      * space. */
{
  if ((addr>=LVM_RAM_BASE)&&(addr<LVM_RAM_BASE+mem_ram_size)) {
    mem_ram[addr-LVM_RAM_BASE]=data;
    return;
  }
  if (~addr<mem_stack_size) {
    mem_stack[~addr]=data;
    return;
  }
}

static lvmw_t
readregs(lvmw_t numb)
     /* Read a LVM register. */
{
  /* This is a private routine: we don't check numb<0x100 */
  return regs[numb];
}

static void
writeregs(lvmw_t numb,lvmw_t data)
     /* Write a LVM register. */
{
  /* This is a private routine: we don't check numb<0x100 */
  if (numb<0x04) return;
  regs[numb]=data;
}

static void
push(lvmw_t w)
     /* Push w on the LVM stack. */
{
  lvmw_t sp;
  mem_stack_size++;
  if (mem_stack_size>mem_stack_all_size)
    mem_stack=lgd_realloc(1,mem_stack,
			  (mem_stack_all_size+=256)*sizeof(lvmw_t));
  sp=readregs(SP_NUM);
  writeregs(SP_NUM,--sp);
  writemem(sp,w);
}

static lvmw_t
pop(void)
     /* Pop a value from the LVM stack. */
{
  lvmw_t sp,w;
  sp=readregs(SP_NUM);
  w=readmem(sp);
  writeregs(SP_NUM,++sp);
  mem_stack_size--;
  return w;
}

static lvmw_t
signextend(lvmw_t w)
     /* Perform sign extension of w: the lower 16 bits are unchanged
      * and the upper 16 bits are copies of the 15th bit. */
{
  if ((w>>15)&1) return (w&0xffff)|0xffff0000;
  else return w&0xffff;
}

static void
setcarry(void)
     /* Set the carry flag. */
{
  writeregs(FL_NUM,readregs(FL_NUM)|1);
}

static void
clrcarry(void)
     /* Clear the carry flag. */
{
  writeregs(FL_NUM,readregs(FL_NUM)&~1);
}

static char
chkcarry(void)
     /* Is the carry flag set? */
{
  return readregs(FL_NUM)&1;
}

static lvmw_t
alu(char type,lvmw_t a,lvmw_t b)
     /* Perform arithmeical or logical operation of type type: see the
      * LVM manual for details. */
{
  lvmw_t c;
  char cf;
  if ((type==1)&&(!chkcarry())) type=0;
  if ((type==3)&&(!chkcarry())) type=2;
  switch (type) {
  case 0: /* ADD */
    c=a+b;
    cf=b>~a;
    break;
  case 1: /* ADC with carry=1 */
    c=a+b+1;
    cf=b>=~a;
    break;
  case 2: /* SUB */
    c=a-b;
    cf=b>a;
    break;
  case 3: /* SBB with carry=1 */
    c=a-b-1;
    cf=b>=a;
    break;
  case 4: return a&b; break;
  case 5: return a|b; break;
  case 6: return ~(a^b); break;
  case 7: return a^b; break;
  default: return 0;
  }
  if (cf) setcarry();
  else clrcarry();
  return c;
}

static char
comp(char type,lvmw_t a,lvmw_t b)
     /* Compare the words a and b according to type: see the LVM
      * manual for details. */
{
  switch (type) {
  case 0: return a==b; break;
  case 1: return a!=b; break;
  case 2: return b|~a; break;
  case 3: return !(a&b); break;
  case 4: return a>=b; break;
  case 5: return a>b; break;
  case 6: return ((int32_t)a)>=((int32_t)b); break;
  case 7: return ((int32_t)a)>((int32_t)b); break;
  }
  return 0;
}

static lvmw_t
shift(char type,lvmw_t w,char count)
     /* Shift w by count bits according to type: see the LVM manual
      * for details. Note that count _must_ be between 0 and 31. */
{
  switch (type) {
  case 0: /* SHR */
    return w>>count;
    break;
  case 1: /* SHL */
    return w<<count;
    break;
  case 2: /* SAR */
    return (lvmw_t)(((int32_t)w)>>count);
    break;
  case 4: /* ROR */
    return (w>>count)|(w<<(32-count));
    break;
  case 5: /* ROL */
    return (w<<count)|(w>>(32-count));
    break;
  }
  return 0;
}

static lvmw_t
immediate(char type,lvmw_t w,lvmw_t i)
     /* Performing of the immediate load operation: see the LVM manual
      * for details. Essentially, this does w=i but unfortunately
      * things aren't that simple. */
{
  switch (type) {
  case 0: /* Set the low-order bits of w to those of i. */
  case 4:
    return (w&0xffff0000)|(i&0xffff);
    break;
  case 1: /* Set the high-order bits of w to the low-order bits of i. */
  case 5:
    return (w&0x0000ffff)|(i<<16);
    break;
  case 2: /* Same as 0 but clear high-order bits. */
    return i&0xffff;
    break;
  case 3: /* Same as 1 but clear low-order bits. */
  case 7:
    return i<<16;
    break;
  case 6: /* Same as 2 but sign-extension instead of zero-extension. */
    return signextend(i);
    break;
  case 8: /* Add i to w. */
  case 10:
    return w+(i&0xffff);
    break;
  case 9: /* Add i<<16 to w. */
  case 11:
  case 13:
  case 15:
    return w+(i<<16);
    break;
  case 12: /* Add i, sign-extended, to w. */
  case 14:
    return w+signextend(i);
    break;
  }
  return 0;
}

static void
service(lvmw_t registry)
     /* Perform a special system service. */
{
  if (registry==0) {
    running=0;
    return;
  }
  if (registry==1) {
    fprintf(stderr,"%#010lx\n",(unsigned long int)readregs(0x81));
  }
}

static void
performstep(void)
     /* Perform one execution step of the LVM. */
{
  lvmw_t pc,instr,w1,w2,w3;
  pc=readregs(PC_NUM);
  instr=readmem(pc);
  writeregs(PC_NUM,++pc);
  switch (instr>>28) {
  case 0x0: /* Arithmetico-logical operation. */
    w1=readregs((instr>>16)&0xff);
    w2=readregs((instr>>8)&0xff);
    w3=alu((instr>>24)&0xf,w1,w2);
    writeregs(instr&0xff,w3);
    break;
  case 0x1: /* Load/store operation. */
    {
      char ldst; /* Whether operation is store. */
      char ind; /* Whether address is indirect. */
      char mul; /* Index scale factor. */
      lvmw_t addr;
      w2=readregs((instr>>8)&0xff);
      w3=readregs(instr&0xff);
      ldst=(instr>>27)&1;
      ind=(instr>>26)&1;
      mul=(instr>>24)&3;
      if (ind) addr=readmem(w2)+(w3<<mul);
      else addr=w2+(w3<<mul);
      if (ldst) writemem(addr,readregs((instr>>16)&0xff));
      else writeregs((instr>>16)&0xff,readmem(addr));
    }
    break;
  case 0x2: /* Comparisons. */
    w1=readregs((instr>>16)&0xff);
    w2=readregs((instr>>8)&0xff);
    w3=comp((instr>>24)&0xf,w1,w2);
    writeregs(instr&0xff,w3);
    break;
  case 0x3: /* Jumps. */
    {
      char immrg; /* Whether destination is a register. */
      char cond; /* Whether condition is regs!=0 rather than regs==0. */
      char rel; /* Whether jump is relative to current pc. */
      char call; /* Whether to save current pc on the stack. */
      lvmw_t addr;
      immrg=(instr>>27)&1;
      cond=(instr>>26)&1;
      rel=(instr>>25)&1;
      call=(instr>>24)&1;
      if (cond==!readregs((instr>>16)&0xff)) break;
      if (call) push(pc);
      if (immrg) addr=readregs(instr&0xff);
      else addr=signextend(instr&0xffff);
      if (rel) addr+=pc;
      else if (immrg) addr=(addr&0xffff)|(pc&0xffff0000);
      writeregs(PC_NUM,addr);
    }
    break;
  case 0x4: /* Shifts. */
    {
      char immrg; /* Whether count is a register. */
      char type; /* Type of shift/rotate. */
      char count;
      immrg=(instr>>27)&1;
      type=(instr>>24)&7;
      if (immrg) count=readregs((instr>>8)&0xff)&31;
      else count=(instr>>8)&31;
      w1=readregs((instr>>16)&0xff);
      w3=shift(type,w1,count);
      writeregs(instr&0xff,w3);
    }
    break;
  case 0x5: /* Immediate load. */
    {
      char type;
      type=(instr>>24)&15;
      w1=readregs((instr>>16)&0xff);
      w1=immediate(type,w1,instr&0xffff);
      writeregs((instr>>16)&0xff,w1);
    }
    break;
  case 0x6: /* Push. */
    w1=readregs(instr&0xff);
    push(w1);
    break;
  case 0x7: /* Pop. */
    w1=pop();
    writeregs(instr&0xff,w1);
    break;
  case 0xf: /* Service call. */
    service(instr&0xffffff);
    break;
  }
}

void
legasm_run(lvmw_t p,lvmw_t q)
     /* Run the legasm program. p and q are parameters, which are
      * passed through registers 0x80 and 0x81. */
{
  running=1;
  writeregs(0x80,p);
  writeregs(0x81,q);
  writeregs(PC_NUM,0);
  writeregs(SP_NUM,0);
  writeregs(FL_NUM,0);
  while (running)
    performstep();
}
