;; This file is part of Legendes II
;; Copyright (c) 1998 by
;;   Philippe Clavel (philippe.clavel@supelec.fr)
;;   David Madore (david.madore@ens.fr)
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation; either version
;; 2 of the License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty
;; of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
;; the GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;

;; file legendesrc.scm

;; This is the guile initialization file.

;;--------------------------------------------------------------------
;; The key bindings defined in this file are as follows:
;;  * the arrow keys move the hero around
;;  * the middle mouse button will teleport him
;;  * the right button will read a map square, and make that the
;;    ``current floor type''
;;  * the left button will apply the current floor type to change
;;    a map square
;;  * the + and - keys will change the floor type (and also make the
;;    new type the current type)
;;  * the control key, with the left and right mouse buttons and with
;;    the + and - keys makes them act on chunks rather than squares
;;  * the space bar serves to mark a position
;;  * meta-left button fills the rectangle from the marked position
;;    to the clicked position with the current floor type
;;  * control-S saves the modifications made to the map and chunks
;;    (that is, it flushes the map and chunk caches - note that if
;;    control-S is not pressed, it is not guaranteed that changes
;;    will be saved, but it is also not guaranteed that they will
;;    _not_ be saved, because the cache can be partially flushed for
;;    lack of space)
;;--------------------------------------------------------------------

(define (notify x y) #t)

;; Displacement keys
(define (key-left x y) (move-hero -1 0))
(define (key-right x y) (move-hero 1 0))
(define (key-up x y) (move-hero 0 -1))
(define (key-down x y) (move-hero 0 1))

(define (move-at-random x y)
  (let ((i (random 4)))
    (if (= i 0) (key-left x y)
      (if (= i 1) (key-right x y)
        (if (= i 2) (key-up x y)
          (if (= i 3) (key-down x y)))))))

;; Map editing section

(define (key-sync-map x y) (sync-map))

(define (mouse-teleport x y)
  (let ((hero-pos (whereis-hero)))
    (move-hero (- x (car hero-pos))
	       (- y (cadr hero-pos)))))

(define current-square-value 0)
(define current-chunk-value 0)
(define current-random-width 1)

(define (get-square-value) current-square-value)
(define (get-chunk-value) current-chunk-value)

(define (set-square-value! v)
  (display "Square value: ") (display current-square-value)
  (set! current-square-value v)
  (display " -> ") (display current-square-value) (newline)
  current-square-value)

(define (set-chunk-value! v)
  (display "Chunk value: ") (display current-chunk-value)
  (set! current-chunk-value v)
  (display " -> ") (display current-chunk-value) (newline)
  current-chunk-value)

(define (set-random-width-1 x y) (set! current-random-width 1))
(define (set-random-width-2 x y) (set! current-random-width 2))
(define (set-random-width-3 x y) (set! current-random-width 3))
(define (set-random-width-4 x y) (set! current-random-width 4))

(define (mouse-put-square x y)
  (write-square x y (+ (get-square-value) (random current-random-width))))
(define (mouse-put-chunk x y) (write-chunk x y (get-chunk-value)))
(define (mouse-get-square x y)
  (set-square-value! (read-square x y))
  (set-random-width-1 x y))
(define (mouse-get-chunk x y) (set-chunk-value! (read-chunk x y)))
(define (mouse-inc-square x y)
  (write-square x y (set-square-value! (+ (read-square x y) 1)))
  (set-random-width-1 x y))
(define (mouse-inc-chunk x y)
  (write-chunk x y (set-chunk-value! (+ (read-chunk x y) 1))))
(define (mouse-dec-square x y)
  (write-square x y (set-square-value! (- (read-square x y) 1)))
  (set-random-width-1 x y))
(define (mouse-dec-chunk x y)
  (write-chunk x y (set-chunk-value! (- (read-chunk x y) 1))))

(define current-mark-x 0)
(define current-mark-y 0)

(define (get-mark) (list current-mark-x current-mark-y))

(define (set-mark! x y)
  (set! current-mark-x x) (set! current-mark-y y))

(define (zone-fill x y)
  (define m (get-mark))
  (define x0 (car m)) (define y0 (cadr m))
  (define x1 x) (define y1 y)
  (define i 0) (define j 0)
  (if (< x1 x0) (list (set! i x0) (set! x0 x1) (set! x1 i)))
  (if (< y1 y0) (list (set! j y0) (set! y0 y1) (set! y1 j)))
  (display "Painting from (") (display x0) (display ",")
  (display y0) (display ") to (") (display x1) (display ",")
  (display y1) (display ").") (newline)
  (set! i x0)
  (while (<= i x1)
	 (set! j y0)
	 (while (<= j y1)
		(mouse-put-square i j)
		(set! j (+ j 1)))
	 (set! i (+ i 1))))

(modify-translations "<Key>Down: notify(key-down)\n\
<Key>Up: notify(key-up)\n\
<Key>Left: notify(key-left)\n\
<Key>Right: notify(key-right)\n\
Ctrl<Key>R: notify(move-at-random)\n\
Ctrl<Btn1Down>: notify(mouse-put-chunk)\n\
Mod1<Btn1Down>: notify(zone-fill)\n\
<Btn1Down>: notify(mouse-put-square)\n\
Ctrl<Btn3Down>: notify(mouse-get-chunk)\n\
<Btn3Down>: notify(mouse-get-square)\n\
Ctrl<Key>plus: notify(mouse-inc-chunk)\n\
<Key>plus: notify(mouse-inc-square)\n\
Ctrl<Key>minus: notify(mouse-dec-chunk)\n\
<Key>minus: notify(mouse-dec-square)\n\
<Key>F1: notify(set-random-width-1)\n\
<Key>F2: notify(set-random-width-2)\n\
<Key>F3: notify(set-random-width-3)\n\
<Key>F4: notify(set-random-width-4)\n\
<Btn2Down>: notify(mouse-teleport)\n\
<Key>space: notify(set-mark!)\n\
Ctrl<Key>s: notify(key-sync-map)")
