/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File output.c */
/* This file contains the output handling thread. */

#include <stdio.h>
#include <pthread.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>

#include "lgd0.h"
#include "fifo.h"
#include "requests.h"

#include "ViewArea.h"
#include "input.h"
#include "icons.h"
#include "pixmaps.h"
#include "output.h"

pthread_t output_thread;

#define WIDTH 32
#define HEIGHT 32

#define DISPEXTRA 2

/* Display coordinates of the upper-left corner */
long int disp_x,disp_y;
/* The serial numbers of the compiled pixmaps being displayed */
unsigned long disp_tab[WIDTH][HEIGHT];

#define MAXSUPERPOSE 16
u_int32_t disp_desc_tab[WIDTH][HEIGHT][MAXSUPERPOSE];

/* disp_redisplay_stat is 1 when the output thread has work to do. */
char disp_redisplay_stat;
/* The following condition serves to wake up the ouput thread when it
 * receives work to do. */
pthread_cond_t disp_redisplay_cond = PTHREAD_COND_INITIALIZER;
/* The following mutex serves to protect the above variable and
 * condition. */
pthread_mutex_t disp_redisplay_mutex = PTHREAD_MUTEX_INITIALIZER;

GC output_gc;

void
demand_redisplay(void)
     /* Wake up the output thread and ask it to redisplay.  This is
      * used by some other thread (a server thread, presumably).
      * _Not_ by the output thread (of course). */
{
  pthread_mutex_lock(&disp_redisplay_mutex);
  disp_redisplay_stat=1;
  pthread_cond_broadcast(&disp_redisplay_cond);
  pthread_mutex_unlock(&disp_redisplay_mutex);
}

void
init_40_output(void)
{
  disp_redisplay_stat=1;
  output_gc=XCreateGC(dpy,disp_pm,0,NULL);
}

static void
append_to_desc(u_int32_t *desc,u_int32_t v)
{
  int i;
  for (i=0;desc[i];i++);
  desc[i++]=v;
  desc[i]=0;
}

static void
append_to_desc_xy(int x,int y,u_int32_t v)
{
  if ((x>=0)&&(y>=0)&&(x<WIDTH)&&(y<HEIGHT))
    append_to_desc(disp_desc_tab[x][y],v);
}

static u_int32_t ground_tiles_tab[] = {
  0x00000000,0x00000001,0x00000002,0x00000003,
  0x00000008,0x00000009,0x0000000a,0x0000000b,
  0x0000000c,0x0000000d,0x0000000e,0x0000000f,
  0x00000020,0x00000021,0x00000022,0x00000023,
  0x00000024,0x00000025,0x00000026,0x00000027,
  0x00000028,0x00000029,0x0000002a,0x0000002b,
  0x0000002c,0x0000002d,0x0000002e,0x0000002f,
};

static u_int32_t hero_tile = 0x01000004;

static u_int32_t extra_tiles_tab[] = {
  0x00000000,0x03000010
};

static void
append_tile(int x,int y,u_int32_t w)
{
  if ((w>>24)==0)
    append_to_desc_xy(x,y,w);
  else if ((w>>24)==1) {
    append_to_desc_xy(x,y,w);
    append_to_desc_xy(x-1,y,w|0x100000);
    append_to_desc_xy(x,y-1,w|0x010000);
    append_to_desc_xy(x-1,y-1,w|0x110000);
  } else if ((w>>24)==3) {
    int i,j;
    for (i=0;i<4;i++) for (j=0;j<4;j++)
      append_to_desc_xy(x-i,y-j,w|(i<<20)|(j<<16));
  }
}

void
output_loop(void)
{
  int i,j;
  long int x,y;
  /* Wait for a redisplay demand */
  pthread_mutex_lock(&disp_redisplay_mutex);
  while (!disp_redisplay_stat)
    pthread_cond_wait(&disp_redisplay_cond,&disp_redisplay_mutex);
  disp_redisplay_stat=0;
  pthread_mutex_unlock(&disp_redisplay_mutex);
  /* Where is the hero? */
  reqq_whereis_hero(&x,&y);
  {
    long int nx,ny,dx,dy;
    nx=x-16;
    ny=y-16;
    dx=nx-disp_x;
    dy=ny-disp_y;
    /* If you are trying to understand this function, this part should
     * probably be the last one you should try to understand.  The
     * whole part from now to the (EOO) comment later on can be
     * removed without ill effect, and only with a performance penalty
     * (and even then, not always).  What it does is it assumes, for
     * optimization purposes, that most of the changes in the display
     * since the last redisplay consist of just moving the display
     * pixmap by an amount equal and opposite to the motion of the
     * hero in that delay.  It therefore displaces the display pixmap
     * and the table of serial numbers accordingly. */
    if ((dx>=0)&&(dy>=0)) {
      for (i=0;i+dx<WIDTH;i++) for (j=0;j+dy<HEIGHT;j++)
	disp_tab[i][j]=disp_tab[i+dx][j+dy];
      XCopyArea(dpy,disp_pm,disp_pm,output_gc,dx*ICON_SIZE,dy*ICON_SIZE,
		(WIDTH-dx)*ICON_SIZE,(HEIGHT-dy)*ICON_SIZE,0,0);
    } else if ((dx>=0)&&(dy<0)) {
      for (i=0;i+dx<WIDTH;i++) for (j=HEIGHT-1;j+dy>=0;j--)
	disp_tab[i][j]=disp_tab[i+dx][j+dy];
      XCopyArea(dpy,disp_pm,disp_pm,output_gc,dx*ICON_SIZE,0,
		(WIDTH-dx)*ICON_SIZE,(HEIGHT+dy)*ICON_SIZE,0,-dy*ICON_SIZE);
    } else if ((dx<0)&&(dy>=0)) {
      for (i=WIDTH-1;i+dx>=0;i--) for (j=0;j+dy<HEIGHT;j++)
	disp_tab[i][j]=disp_tab[i+dx][j+dy];
      XCopyArea(dpy,disp_pm,disp_pm,output_gc,0,dy*ICON_SIZE,
		(WIDTH+dx)*ICON_SIZE,(HEIGHT-dy)*ICON_SIZE,-dx*ICON_SIZE,0);
    } else if ((dx<0)&&(dy<0)) {
      for (i=WIDTH-1;i+dx>=0;i--) for (j=HEIGHT-1;j+dy>=0;j--)
	disp_tab[i][j]=disp_tab[i+dx][j+dy];
      XCopyArea(dpy,disp_pm,disp_pm,output_gc,0,0,
		(WIDTH+dx)*ICON_SIZE,(HEIGHT-dy)*ICON_SIZE,
		-dx*ICON_SIZE,-dy*ICON_SIZE);
    }
    /* (EOO) */
    disp_x=nx;
    disp_y=ny;
  }
  /* Now we fill in the description array */
  for (i=0;i<WIDTH;i++) for (j=0;j<HEIGHT;j++)
    disp_desc_tab[i][j][0]=0;
  for (i=0;i<WIDTH+DISPEXTRA;i++) for (j=0;j<HEIGHT+DISPEXTRA;j++) {
    u_int32_t v;
    char h;
    x=disp_x+i;
    y=disp_y+j;
    reqq_readmap(x,y,&v,&h);
    append_tile(i,j,ground_tiles_tab[v&0xffff]);
    if (h)
      append_tile(i,j,hero_tile);
    append_tile(i,j,extra_tiles_tab[v>>16]);
  }
  /* Now we draw the pixmap */
  for (i=0;i<WIDTH;i++) for (j=0;j<HEIGHT;j++) {
    Pixmap pm;
    unsigned long int serial;
    pm=get_pixmap(disp_desc_tab[i][j],&serial);
    if (serial!=disp_tab[i][j]) {
      XCopyArea(dpy,pm,disp_pm,output_gc,0,0,ICON_SIZE,ICON_SIZE,
		i*ICON_SIZE,j*ICON_SIZE);
      disp_tab[i][j]=serial;
    }
  }
  demand_redraw();
}

void *
output_main(void *garbage)
{
  while (1)
    output_loop();
}
