/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File pnm2lgi.c */
/* This file contains a standalone program that will convert a
 * Portable aNyMap to a LeGendes Icon file. */

#include <stdio.h>
#include <string.h>

#include "lgd0.h"
#include "config.h"
#include "lgdlib.h"

char *
next_word(FILE *f)
     /* Read the next word from the PNM file f, skipping whitespace if
      * necessary. */
{
  char *s;
  int len,maxlen;
  int ch;
  char comment,skip;
  s=NULL;
  /* Skip initial whitespace and comments. */
  comment=0;
  do {
    ch=getc(f);
    if (ch==EOF) return NULL;
    if (comment) {
      if (ch=='\n') { comment=0; skip=1; }
      else skip=1;
    } else {
      if (ch=='#') { comment=1; skip=1; }
      else if ((ch==' ')||(ch=='\n')||(ch=='\r')||(ch=='\011')) skip=1;
      else skip=0;
    }
  } while (skip);
  /* Read the word. */
  maxlen=0;
  len=1;
  do {
    ++len;
    while (len>maxlen)
      s=lgd_realloc(1,s,maxlen+=16);
    s[len-2]=ch;
    s[len-1]=0;
    ch=getc(f);
    if (ch==EOF) return s;
    if (ch=='#') { comment=1; skip=1; }
    else if ((ch==' ')||(ch=='\n')||(ch=='\r')||(ch=='\011')) skip=1;
    else skip=0;
  } while (!skip);
  /* Skip final comment(s) if any. */
  while (comment) {
    ch=getc(f);
    if (ch==EOF) return s;
    if (ch=='\n') { comment=0; }
  }
  return s;
}

unsigned int
next_number(FILE *f)
     /* Read the next word from the PNM file f; it must be a
      * number. */
{
  char *s;
  unsigned int n;
  s=next_word(f);
  n=0;
  sscanf(s,"%i",&n);
  lgd_free(s);
  return n;
}

#define ICON_SIZE 16

#define MAX_SIZE (ICON_SIZE*4)

FILE *colorf,*transf;

struct blabla_s {
  unsigned int r,g,b;
} color[MAX_SIZE][MAX_SIZE],trans[MAX_SIZE][MAX_SIZE];

int icon_n;

int size; /* 0 if not decided yet */
int size_in_icons;

lgd_filedesc_t outf;

typedef struct color_s {
  unsigned int r,g,b,a; /* Range 0-255 (even though unsigned int) */
} color_t;

typedef color_t icon_t[ICON_SIZE][ICON_SIZE];

icon_t icon;

u_int32_t tab[ICON_SIZE][ICON_SIZE];

void
readpnm(FILE *f,struct blabla_s data[][MAX_SIZE])
     /* Read the PNM file f and store its data in the data tab.  Also
      * set the global variable size if not done already. */
{
  int i,j;
  char format;
  if (f==NULL) {
    for (i=0;i<MAX_SIZE;i++) for (j=0;j<MAX_SIZE;j++) {
      data[i][j].r=0;
      data[i][j].g=0;
      data[i][j].b=0;
    }
  } else {
    char *s;
    int w,h,max;
    s=next_word(f);
    if (strcmp(s,"P1")==0) format=1;
    else if (strcmp(s,"P2")==0) format=2;
    else if (strcmp(s,"P3")==0) format=3;
    else lgd_fatal("Unknown PNM format %s",s);
    lgd_free(s);
    w=next_number(f);
    if (!size) {
      if (w&0xf) lgd_fatal("Width %d is not multiple of 16",w);
      if (w>MAX_SIZE) lgd_fatal("Width %d is too big");
      size=w;
      fprintf(stderr,"(Size will be %d)\n",size);
    } else if (w!=size) lgd_fatal("Width %d does not match size %d",w,size);
    h=next_number(f);
    if (h!=size) lgd_fatal("Height %d does not match size %d",h,size);
    if ((format==2)||(format==3))
      max=next_number(f);
    else max=1;
    for (i=0;i<size;i++) for (j=0;j<size;j++) {
      if (format==3) {
	data[i][j].r=(next_number(f)*255)/max;
	data[i][j].g=(next_number(f)*255)/max;
	data[i][j].b=(next_number(f)*255)/max;
      } else {
	data[i][j].r=(next_number(f)*255)/max;
	data[i][j].g=data[i][j].r;
	data[i][j].b=data[i][j].r;
      }
    }
  }
}

int
main(int argc,char **argv)
{
  if (argc<3) {
    fprintf(stderr,"Syntax is %s <output file> <icon number>\n");
    fprintf(stderr,"     [color input file [transparency input file]]\n");
    fprintf(stderr,
	    "Replace the name of an input file by - for standard input\n");
    fprintf(stderr,"And by . for just pure black.\n");
    exit(0);
  }
  lgd_open(1,&outf,argv[1],OPEN_READWRITE);
  icon_n=0;
  sscanf(argv[2],"%i",&icon_n);
  if (argc>=4) {
    if (strcmp(argv[3],"-")==0)
      colorf=stdin;
    else if ((colorf=fopen(argv[3],"r"))==NULL) {
      perror("Opening color input file");
      exit(1);
    }
  } else colorf=stdin;
  if (argc>=5) {
    if (strcmp(argv[4],"-")==0)
      transf=stdin;
    else if (strcmp(argv[4],".")==0)
      transf=NULL;
    else if ((transf=fopen(argv[4],"r"))==NULL) {
      perror("Opening transparency input file");
      exit(1);
    }
  } else transf=NULL;
  readpnm(colorf,color);
  readpnm(transf,trans);
  if (!size) size=16;
  size_in_icons=size/16;
  {
    int u,v;
    for (u=0;u<size_in_icons;u++) for (v=0;v<size_in_icons;v++) {
      int i,j;
      for (i=0;i<ICON_SIZE;i++) for (j=0;j<ICON_SIZE;j++) {
	icon[i][j].r=
	  color[j+(size_in_icons-v-1)*16][i+(size_in_icons-u-1)*16].r;
	icon[i][j].g=
	  color[j+(size_in_icons-v-1)*16][i+(size_in_icons-u-1)*16].g;
	icon[i][j].b=
	  color[j+(size_in_icons-v-1)*16][i+(size_in_icons-u-1)*16].b;
	icon[i][j].a=
	  trans[j+(size_in_icons-v-1)*16][i+(size_in_icons-u-1)*16].r;
	tab[i][j]=(icon[i][j].r)|(icon[i][j].g<<8)|(icon[i][j].b<<16)|
	  (icon[i][j].a<<24);
      }
      fprintf(stderr,"Writing square (%d,%d) as icon number %d\n",u,v,icon_n);
      lgd_write_littleendian_int32(1,outf,(u_int32_t *)tab,
				   (icon_n++)*ICON_SIZE*ICON_SIZE,
				   ICON_SIZE*ICON_SIZE);
    }
    lgd_close(1,outf);
  }
}
