/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File requests.c */
/* This file contains routines used by the client threads to submit
 * requests in the control fifo. */

#include <stdio.h>
#include <pthread.h>

#include "lgd0.h"
#include "config.h"
#include "lgdlib.h"
#include "fifo.h"
#include "requests.h"

/* The following req_ functions all work in the same way: you pass to
 * them a pointer to a msg_response_s structure in which you want the
 * answer to be put (NULL if you don't want to hear about the request
 * any more), and optional parameters to the call.  The function will
 * allocate an msg_fifo_s structure, fill in the entries for you, and
 * append that structure to the control fifo (see <fifo.h>).  Note
 * that the answer is not available immediately: use wait_for_answer()
 * when you are ready to hear it, or use the reqq_ routines for this
 * (if you have nothing constructive to do between placing the
 * question and getting the answer). */

void
req_noop(msg_response_t *resp)
     /* Submit a noop request.  This is not very interesting, but it
      * could be used for testing purposes.  Also, copy this function
      * when you want to add a new request handling routine. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_noop;
  append_control_msg(msg);
}

void
req_die(msg_response_t *resp)
     /* Submit a die request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_die;
  append_control_msg(msg);
}

void
req_readmap(msg_response_t *resp,long int x,long int y)
     /* Submit a readmap request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_readmap;
  msg->param.readmap.x=x;
  msg->param.readmap.y=y;
  append_control_msg(msg);
}

void
req_write_chunk(msg_response_t *resp,long int x,long int y,u_int32_t v)
     /* Submit a write_chunk request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_write_chunk;
  msg->param.write_chunk.x=x;
  msg->param.write_chunk.y=y;
  msg->param.write_chunk.v=v;
  append_control_msg(msg);
}

void
req_write_square(msg_response_t *resp,long int x,long int y,u_int32_t v)
     /* Submit a write_square request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_write_square;
  msg->param.write_square.x=x;
  msg->param.write_square.y=y;
  msg->param.write_square.v=v;
  append_control_msg(msg);
}

void
req_read_chunk(msg_response_t *resp,long int x,long int y)
     /* Submit a read_chunk request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_read_chunk;
  msg->param.read_chunk.x=x;
  msg->param.read_chunk.y=y;
  append_control_msg(msg);
}

void
req_read_square(msg_response_t *resp,long int x,long int y)
     /* Submit a read_square request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_read_square;
  msg->param.read_square.x=x;
  msg->param.read_square.y=y;
  append_control_msg(msg);
}

void
req_sync_map(msg_response_t *resp)
     /* Submit a sync_map request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_sync_map;
  append_control_msg(msg);
}

void
req_move_hero(msg_response_t *resp,long int dx,long int dy)
     /* Submit a move_hero request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_move_hero;
  msg->param.move_hero.dx=dx;
  msg->param.move_hero.dy=dy;
  append_control_msg(msg);
}

void
req_whereis_hero(msg_response_t *resp)
     /* Submit a whereis_hero request. */
{
  msg_fifo_t *msg;
  msg=lgd_malloc(1,sizeof(msg_fifo_t));
  msg->in_heap=1;
  msg->response=resp;
  msg->type=msg_whereis_hero;
  append_control_msg(msg);
}

void
wait_for_answer(msg_response_t *resp)
     /* Sleep until the answer has been filled. */
{
  pthread_mutex_lock(&resp->ready_mutex);
  while (!resp->ready_stat)
    pthread_cond_wait(&resp->ready_cond,&resp->ready_mutex);
  pthread_mutex_unlock(&resp->ready_mutex);
}

/* The reqq_ functions provide more a more convenient interface to the
 * server requests.  However, they use wait_for_answer(), so it is not
 * always a good idea to use them (the benefit of multithreading may
 * be lost.  Another difference is that the msg_fifo_s structure they
 * pass to the control fifo is found on their stack rather than being
 * allocated on the heap, which is more efficient since it doesn't
 * incur the malloc() time penalty (but the req_ functions cannot do
 * the same since the function might terminate before the request is
 * treated, which means the stack contents would be zapped). */

msg_errcode_t
reqq_noop(void)
     /* Submit a noop request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_noop;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  return resp.errcode;
}

msg_errcode_t
reqq_die(void)
     /* Submit a die request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_die;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  return resp.errcode;
}

msg_errcode_t
reqq_readmap(long int x,long int y,u_int32_t *val,char *hero)
     /* Submit a readmap request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_readmap;
  msg.param.readmap.x=x;
  msg.param.readmap.y=y;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  if (val) *val=resp.ret.readmap.val;
  if (hero) *hero=resp.ret.readmap.hero;
  return resp.errcode;
}

msg_errcode_t
reqq_write_chunk(long int x,long int y,u_int32_t v)
     /* Submit a write_chunk request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_write_chunk;
  msg.param.write_chunk.x=x;
  msg.param.write_chunk.y=y;
  msg.param.write_chunk.v=v;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  return resp.errcode;
}

msg_errcode_t
reqq_write_square(long int x,long int y,u_int32_t v)
     /* Submit a write_square request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_write_square;
  msg.param.write_square.x=x;
  msg.param.write_square.y=y;
  msg.param.write_square.v=v;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  return resp.errcode;
}

msg_errcode_t
reqq_read_chunk(long int x,long int y,u_int32_t *val)
     /* Submit a read_chunk request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_read_chunk;
  msg.param.read_chunk.x=x;
  msg.param.read_chunk.y=y;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  if (val) *val=resp.ret.read_chunk.val;
  return resp.errcode;
}

msg_errcode_t
reqq_read_square(long int x,long int y,u_int32_t *val)
     /* Submit a read_square request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_read_square;
  msg.param.read_square.x=x;
  msg.param.read_square.y=y;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  if (val) *val=resp.ret.read_square.val;
  return resp.errcode;
}

msg_errcode_t
reqq_sync_map(void)
     /* Submit a sync_map request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_sync_map;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  return resp.errcode;
}

msg_errcode_t
reqq_move_hero(long int dx,long int dy)
     /* Submit a move_hero request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_move_hero;
  msg.param.move_hero.dx=dx;
  msg.param.move_hero.dy=dy;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  return resp.errcode;
}

msg_errcode_t
reqq_whereis_hero(long int *x,long int *y)
     /* Submit a whereis_hero request and wait for answer */
{
  msg_fifo_t msg;
  msg_response_t resp;
  msg.in_heap=0;
  msg.response=&resp;
  msg.type=msg_whereis_hero;
  append_control_msg(&msg);
  wait_for_answer(&resp);
  if (x) *x=resp.ret.whereis_hero.x;
  if (y) *y=resp.ret.whereis_hero.y;
  return resp.errcode;
}
